from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization


def generate_key():
    return rsa.generate_private_key(public_exponent=65537, key_size=4096)


def write_key(key, path, password=None):
    if password:
        password = password.encode('utf-8')
        encryption_algorithm = serialization.BestAvailableEncryption(password) 
    else:
        encryption_algorithm = serialization.NoEncryption()
    pem = key.private_bytes(encoding=serialization.Encoding.PEM, format=serialization.PrivateFormat.PKCS8, encryption_algorithm=encryption_algorithm)
    with open(path, 'wb') as file:
        file.write(pem)


def load_key(path, password=None):
    if password:
        password = password.encode('utf-8')
    with open(path, "rb") as file:
        return serialization.load_pem_private_key(file.read(), password)


from satispaython.core import send_request
from cryptography.hazmat.primitives import serialization


def obtain_key_id(key, token, staging=False):
    target = '/g_business/v1/authentication_keys'
    public_pem = key.public_key().public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)
    public_pem = public_pem.decode('utf-8')
    body = { 'public_key': public_pem, 'token': token }
    return send_request(None, None, 'post', target, body, None, staging)


def test_authentication(key_id, key):
    target = '/wally-services/protocol/tests/signature'
    body = { 'flow': 'MATCH_CODE', 'amount_unit': 100, 'currency': 'EUR' }
    return send_request(key_id, key, 'post', target, body, None, True)


def create_payment(key_id, key, amount_unit, currency, callback_url, expiration_date=None, external_code=None, metadata=None, idempotency_key=None, staging=False):
    target = '/g_business/v1/payments'
    body = { 'flow': 'MATCH_CODE', 'amount_unit': amount_unit, 'currency': currency, 'callback_url': callback_url }
    if external_code:
        body = { **body, 'external_code': external_code }
    if metadata:
        body = { **body, 'metadata': metadata }
    if expiration_date:
        body = { **body, 'expiration_date': expiration_date }
    return send_request(key_id, key, 'post', target, body, idempotency_key, staging)


def get_payment_details(key_id, key, payment_id, staging=False):
    target = f'/g_business/v1/payments/{payment_id}'
    return send_request(key_id, key, 'get', target, None, None, staging)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="satispaython",
    version="0.1.3",
    author="Daniele Pira",
    author_email="daniele.pira@otto.to.it",
    description="A simple library to manage Satispay payments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/otto-torino/satispaython",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
    install_requires=[
       'requests>=2.24',
       'cryptography>=3.1'
    ],
    python_requires='>=3.8',
)


import requests, json

from satispaython.exceptions import UnexpectedRequestMethod
from base64 import b64encode
from hashlib import sha256
from datetime import datetime, timezone
from cryptography.hazmat.primitives.asymmetric import padding as paddings
from cryptography.hazmat.primitives import hashes


def get_formatted_date():
    return datetime.now(timezone.utc).strftime('%a, %d %b %Y %H:%M:%S %z')


def compute_digest(body):
    body = body.encode('utf-8')
    digest = b64encode(sha256(body).digest())
    digest = digest.decode('utf-8')
    return f'SHA-256={digest}' 


def generate_string(method, target, host, date, digest):
    return f'(request-target): {method} {target}\nhost: {host}\ndate: {date}\ndigest: {digest}'


def sign_string(key, string):
    string = string.encode('utf-8')
    padding = paddings.PSS(mgf=paddings.MGF1(hashes.SHA256()), salt_length=paddings.PSS.MAX_LENGTH)
    signature = key.sign(data=string, padding=paddings.PKCS1v15(), algorithm=hashes.SHA256())
    return b64encode(signature).decode('utf-8')


def compute_authorization_header(key_id, signature):
    return f'Signature keyId="{key_id}", algorithm="rsa-sha256", headers="(request-target) host date digest", signature="{signature}"'
    

def compute_authorization_headers(key, key_id, method, target, host, body):
    date = get_formatted_date()
    digest = compute_digest(body)
    string = generate_string(method, target, host, date, digest)
    signature = sign_string(key, string)
    header = compute_authorization_header(key_id, signature)
    return { 'Host': host, 'Date': date, 'Digest': digest, 'Authorization': header }


def generate_headers(key, key_id, method, target, host, body, idempotency_key):
    headers = { 'Accept': 'application/json' }
    if method == 'post' or method == 'put':
        headers = { **headers, 'Content-Type': 'application/json' }
    if idempotency_key:
        headers = { **headers, 'Idempotency-Key': idempotency_key }
    if key and key_id:
        authorization_headers = compute_authorization_headers(key, key_id, method, target, host, body)
        headers = { **headers, **authorization_headers }
    return headers


def send_request(key_id, key, method, target, body, idempotency_key, staging):
    host = 'staging.authservices.satispay.com' if staging == True else 'authservices.satispay.com'
    body = json.dumps(body) if body else ''
    headers = generate_headers(key, key_id, method, target, host, body, idempotency_key)
    url = f'https://{host}{target}'
    if method == 'get':
        return requests.get(url=url, data=body, headers=headers)
    elif method == 'post':
        return requests.post(url=url, data=body, headers=headers)
    else:
        raise exceptions.UnexpectedRequestMethod('The request method is invalid (should be "get" or "post")')
